package s3rt

import (
	"bytes"
	"context"
	"crypto/md5"
	"flag"
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"os"
	"strconv"
	"sync"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"

	"github.com/aws/aws-sdk-go/aws"
	"github.com/aws/aws-sdk-go/aws/awserr"
	awsRequest "github.com/aws/aws-sdk-go/aws/request"
	"github.com/aws/aws-sdk-go/aws/session"
	"github.com/aws/aws-sdk-go/service/s3"
	"github.com/aws/aws-sdk-go/service/s3/s3iface"
	"github.com/aws/aws-sdk-go/service/s3/s3manager"
	"github.com/aws/aws-sdk-go/service/s3/s3manager/s3manageriface"
)

var (
	ramBuffers = true
	userAgent  = "OIO-S3RT"
	mpuSize    = 5 * 1024 * 1024
	size       = mpuSize - 1
	bucketRT   = false
	timeout    = 10 * time.Second
	interval   = 10 * time.Second
)

func InitFlags(fs *flag.FlagSet) {
	fs.IntVar(&mpuSize, "s3rt.mpu_size", mpuSize, "[s3rt] MPU size")
	fs.IntVar(&size, "s3rt.size", size, "[s3rt] Put object size")
	fs.BoolVar(&ramBuffers, "s3rt.ram_buffers", ramBuffers, "[s3rt] Allocate data in RAM instead of using kernel devices")
	fs.BoolVar(&bucketRT, "s3rt.bucket_rt", bucketRT, "[s3rt] Perform a bucket roundtrip")
	fs.DurationVar(&timeout, "s3rt.timeout", timeout, "[s3rt] operation timeout")
	fs.DurationVar(&interval, "s3rt.interval", interval, "[s3rt] interval between polls")
	fs.StringVar(&userAgent, "s3rt.user_agent", userAgent, "[s3rt] userAgent identifier to add to requests")
}

var log = _log.New("s3rt")

type s3c struct {
	s3        s3iface.S3API
	bucketRT  bool
	bucket    string
	objects   map[string]*object
	service   *common.Service
	timeout   time.Duration
	src       io.ReadCloser
	dst       io.Writer
	dl        s3manageriface.DownloaderAPI
	ul        s3manageriface.UploaderAPI
	userAgent awsRequest.Option
}

type s3rtDaemon struct {
	daemon.Daemon

	clients       []*s3c
	bucketCreated []bool
	errors        []*common.Service
	timeout       time.Duration
	namespace     string
}

type object struct {
	name string
	size int64
}

type FakeWriterAt struct {
	w io.Writer
}

func (fw FakeWriterAt) WriteAt(p []byte, offset int64) (n int, err error) {
	return fw.w.Write(p)
}

func max(a, b int) int {
	if a > b {
		return a
	}
	return b
}

func makeInputOutput(ramBuffers bool, size int) (io.ReadCloser, io.Writer, error) {
	if ramBuffers {
		buf := bytes.NewBuffer(make([]byte, size))
		return ioutil.NopCloser(buf), buf, nil
	}
	fd, err := os.Open("/dev/zero")
	if err != nil {
		log.Warnln("error while opening I/O device", err)
		return nil, nil, err
	}
	return fd, ioutil.Discard, nil
}

func MakeDaemon(namespace string) *s3rtDaemon {
	return &s3rtDaemon{
		clients:   []*s3c{},
		errors:    []*common.Service{},
		timeout:   timeout,
		namespace: namespace,
	}
}

func (d *s3rtDaemon) ID() string {
	return "S3rt"
}

func (d *s3rtDaemon) Interval() time.Duration {
	return interval
}

func (d *s3rtDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	d.reload(services)

	return nil
}

func (d *s3rtDaemon) Close(ctx context.Context) error {
	for _, c := range d.clients {
		c.cleanup()
	}
	return nil
}

func (d *s3rtDaemon) OnServicesChange(services common.ServicesMap) error {
	d.reload(services)

	return nil
}

func (d *s3rtDaemon) reload(services common.ServicesMap) {
	swiftServices := services["oioswift"]

	clients := make([]*s3c, 0, len(swiftServices))
	errServices := make([]*common.Service, 0, len(swiftServices))
	for _, svc := range swiftServices {
		endpoint := fmt.Sprintf("http://%s:%d", svc.IP, svc.Port)
		config := aws.Config{
			Endpoint:         aws.String(endpoint),
			S3ForcePathStyle: aws.Bool(true),
			MaxRetries:       aws.Int(1),
			HTTPClient: &http.Client{
				Timeout: d.timeout,
				Transport: &http.Transport{
					// Prevent 100-Continue
					ExpectContinueTimeout: 0,
				},
			},
		}

		// Ensure to load config and credentials from ~/.aws
		sess, err := session.NewSessionWithOptions(session.Options{
			Config:            config,
			SharedConfigState: session.SharedConfigEnable,
		})
		if err != nil {
			log.Warnln("error while initializing client", svc, err)
			continue
		}

		in, out, err := makeInputOutput(ramBuffers, max(mpuSize+1, size))
		s3Client := s3.New(sess)
		if err != nil {
			log.Warnln("error while initializing client", svc, err)
			continue
		}
		ua := awsRequest.WithAppendUserAgent(userAgent)

		clients = append(clients, &s3c{
			s3:       s3Client,
			bucketRT: bucketRT,
			bucket:   fmt.Sprintf("s3rt-%x", md5.Sum([]byte(endpoint))),
			objects: map[string]*object{
				"simple": &object{name: "s3roundtrip_object_simple", size: int64(size)},
				"ttfb":   &object{name: "s3roundtrip_ttfb_object", size: int64(1)},
				"mpu":    &object{name: "s3roundtrip_mpu_object", size: int64(mpuSize + 1)},
			},
			service: svc,
			timeout: d.timeout,
			src:     in,
			dst:     out,
			dl: s3manager.NewDownloaderWithClient(s3Client, func(d *s3manager.Downloader) {
				d.RequestOptions = append(d.RequestOptions, ua)
			}),
			ul: s3manager.NewUploaderWithClient(s3Client, func(u *s3manager.Uploader) {
				u.PartSize = int64(mpuSize)
				u.RequestOptions = append(u.RequestOptions, ua)
			}),
			userAgent: ua,
		})
	}

	d.bucketCreated = make([]bool, len(clients))

	for _, c := range d.clients {
		c.cleanup()
	}

	d.clients = clients
	d.errors = errServices
}

func (d *s3rtDaemon) Collect(m *common.MetricData) error {
	// TODO: do this only once?
	for _, svc := range d.errors {
		m.IncrError("s3rt", svc.Type)
	}

	var wg sync.WaitGroup
	for i, c := range d.clients {
		wg.Add(1)
		// Concurrent writing to different indices of a same slice is safe
		// ...but just to be sure
		go func(c *s3c, bucketCreated *bool) {
			defer wg.Done()

			var t time.Duration
			var err error

			if !*bucketCreated || c.bucketRT {
				t, err = c.mb(c.bucket)
				if err == nil {
					*bucketCreated = true
				} else if aerr, ok := err.(awserr.Error); ok && aerr.Code() == s3.ErrCodeBucketAlreadyExists {
					*bucketCreated = true
				} else if !c.bucketRT {
					m.IncrError("s3rt", "s3rt")
					log.Warnln("error while precreating bucket", c, err)
				}
				if c.bucketRT {
					d.s3Publish(m, c.service, "mb", t, err, "")
				}
			}

			for type_, obj := range c.objects {
				t, err := c.put(c.bucket, obj)
				if err == nil {
					t, err = c.get(c.bucket, obj)
					d.s3Publish(m, c.service, "get", t, err, type_)
					t, err = c.del(c.bucket, obj)
					d.s3Publish(m, c.service, "del", t, err, type_)
				}
				d.s3Publish(m, c.service, "put", t, err, type_)
			}

			t, err = c.ls(c.bucket)
			d.s3Publish(m, c.service, "ls", t, err, "")

			if c.bucketRT {
				t, err = c.rb(c.bucket)
				d.s3Publish(m, c.service, "rb", t, err, "")
			}
		}(c, &d.bucketCreated[i])
	}
	wg.Wait()

	return nil
}

func (d *s3rtDaemon) s3Publish(m *common.MetricData, service *common.Service, request string, t time.Duration, err error, type_ string) {
	code := "2xx"
	if err != nil {
		if req, ok := err.(awserr.RequestFailure); ok {
			// Error is an AWS request failure
			status := req.StatusCode() / 100
			if status == 3 || status == 1 {
				// Treat redirects and continues as 2xx
				status = 2
			}
			code = strconv.Itoa(status) + "xx"
		} else {
			code = "other"
		}
	}
	labels := [][]string{
		[]string{"namespace", d.namespace},
		[]string{"service", service.ID},
		[]string{"service_type", service.Type},
		[]string{"service_addr", fmt.Sprintf("%s:%d", service.IP, service.Port)},
		[]string{"method", request},
		[]string{"code", code},
	}
	if type_ != "" {
		labels = append(labels, []string{"type", type_})
	}

	m.Set(&common.Metric{
		Name:   "s3rt_response_time_seconds",
		Labels: labels,
		Value:  float64(t.Seconds()),
	})
}

func (c *s3c) cleanup() {
	c.src.Close()
}

func (c *s3c) put(bucket string, obj *object) (time.Duration, error) {
	input := &s3manager.UploadInput{
		Bucket: aws.String(bucket),
		Key:    aws.String(obj.name),
		Body:   io.LimitReader(c.src, obj.size),
	}
	ctx, cancel := context.WithTimeout(context.Background(), c.timeout)
	defer cancel()
	start := time.Now()
	_, err := c.ul.UploadWithContext(ctx, input)
	return time.Since(start), err
}

func (c *s3c) get(bucket string, obj *object) (time.Duration, error) {
	input := &s3.GetObjectInput{
		Bucket: aws.String(bucket),
		Key:    aws.String(obj.name),
	}
	ctx, cancel := context.WithTimeout(context.Background(), c.timeout)
	defer cancel()
	start := time.Now()
	_, err := c.dl.DownloadWithContext(ctx, FakeWriterAt{w: c.dst}, input)
	return time.Since(start), err
}

func (c *s3c) del(bucket string, obj *object) (time.Duration, error) {
	input := &s3.DeleteObjectInput{
		Bucket: aws.String(bucket),
		Key:    aws.String(obj.name),
	}
	ctx, cancel := context.WithTimeout(context.Background(), c.timeout)
	defer cancel()
	start := time.Now()
	_, err := c.s3.DeleteObjectWithContext(ctx, input, c.userAgent)
	return time.Since(start), err
}

func (c *s3c) mb(bucket string) (time.Duration, error) {
	input := &s3.CreateBucketInput{
		Bucket: aws.String(bucket),
	}
	ctx, cancel := context.WithTimeout(context.Background(), c.timeout)
	defer cancel()
	start := time.Now()
	_, err := c.s3.CreateBucketWithContext(ctx, input, c.userAgent)
	if aerr, ok := err.(awserr.Error); ok && aerr.Code() == s3.ErrCodeBucketAlreadyOwnedByYou {
		return time.Since(start), nil
	}
	return time.Since(start), err
}

func (c *s3c) rb(bucket string) (time.Duration, error) {
	input := &s3.DeleteBucketInput{
		Bucket: aws.String(bucket),
	}
	ctx, cancel := context.WithTimeout(context.Background(), c.timeout)
	defer cancel()
	start := time.Now()
	_, err := c.s3.DeleteBucketWithContext(ctx, input, c.userAgent)
	return time.Since(start), err
}

func (c *s3c) ls(bucket string) (time.Duration, error) {
	input := &s3.ListObjectsInput{
		Bucket:  aws.String(bucket),
		MaxKeys: aws.Int64(1000),
	}
	ctx, cancel := context.WithTimeout(context.Background(), c.timeout)
	defer cancel()
	start := time.Now()
	_, err := c.s3.ListObjectsWithContext(ctx, input, c.userAgent)
	return time.Since(start), err
}
